//#####################################################################
// Copyright 2002-2007, Robert Bridson, Ronald Fedkiw, Geoffrey Irving, Sergey Koltakov, Nipun Kwatra, Andrew Selle, Jonathan Su, Joseph Teran.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class SEGMENT_3D  
//##################################################################### 
#ifndef __SEGMENT_3D__
#define __SEGMENT_3D__

#include <PhysBAM_Tools/Arrays/ARRAYS_FORWARD.h>
#include <PhysBAM_Tools/Vectors/VECTOR_3D.h>
namespace PhysBAM{

template<class T>
class SEGMENT_3D
{
    typedef VECTOR<T,3> TV;
public:
    VECTOR<T,3> x1,x2;

    SEGMENT_3D()
        :x1(0,0,0),x2(1,0,0)
    {}

    SEGMENT_3D(const VECTOR<T,3>& x1_input,const VECTOR<T,3>& x2_input)
        :x1(x1_input),x2(x2_input)
    {}

    template<class T_ARRAY>
    SEGMENT_3D(const T_ARRAY& X_input)
        :x1(X_input(1)),x2(X_input(2))
    {
        STATIC_ASSERT(T_ARRAY::m==2);
    }

    T Length() const
    {return (x2-x1).Magnitude();}

    T Size() const
    {return Length();}

    static T Size(const TV& x1,const TV& x2)
    {return (x2-x1).Magnitude();}

    template<class T_ARRAY>
    static T Signed_Size(const T_ARRAY& X)
    {STATIC_ASSERT(T_ARRAY::m==2);return Size(X(1),X(2));}

    template<class T_ARRAY>
    static T Size(const T_ARRAY& X)
    {return Signed_Size(X);}

    static VECTOR<T,3> Normal(const VECTOR<T,3>& x1,const VECTOR<T,3>& x2,const VECTOR<T,3>& x3) 
    {VECTOR<T,3> v=x2-x1,face_normal=VECTOR<T,3>::Cross_Product(v,x3-x1);
    return VECTOR<T,3>::Cross_Product(face_normal,v).Normalized();} // rotate by 90 degrees clockwise

    static VECTOR<T,3> Normal_Direction(const VECTOR<T,3>& x1,const VECTOR<T,3>& x2,const VECTOR<T,3>& x3) 
    {VECTOR<T,3> v=x2-x1,face_normal=VECTOR<T,3>::Cross_Product(v,x3-x1);
    return VECTOR<T,3>::Cross_Product(face_normal,v);} // can have any magnitude

//#####################################################################
    VECTOR<T,3> Closest_Point_On_Segment(const VECTOR<T,3>& point) const;
    T Distance_From_Point_To_Segment(const VECTOR<T,3>& point) const;
    VECTOR<T,3> Closest_Point_On_Line(const VECTOR<T,3>& point) const;
    T Distance_From_Point_To_Line(const VECTOR<T,3>& point) const;
    VECTOR<T,3> Shortest_Vector_Between_Lines(const SEGMENT_3D<T>& segment,VECTOR<T,2>& weights) const;
    VECTOR<T,3> Shortest_Vector_Between_Segments(const SEGMENT_3D<T>& segment,VECTOR<T,2>& weights) const;
    T Interpolation_Fraction(const VECTOR<T,3>& location) const;
    VECTOR<T,2> Barycentric_Coordinates(const VECTOR<T,3>& location) const;
    VECTOR<T,2> Clamped_Barycentric_Coordinates(const VECTOR<T,3>& location,const T tolerance=1e-7) const;
    static T Interpolation_Fraction(const VECTOR<T,3>& location,const VECTOR<T,3>& x1,const VECTOR<T,3>& x2);
    static VECTOR<T,2> Barycentric_Coordinates(const VECTOR<T,3>& location,const VECTOR<T,3>& x1,const VECTOR<T,3>& x2);
    static VECTOR<T,2> Clamped_Barycentric_Coordinates(const VECTOR<T,3>& location,const VECTOR<T,3>& x1,const VECTOR<T,3>& x2,const T tolerance=1e-7);
//#####################################################################
};   
}
#endif
